# -*- Mode:Python; coding: utf-8 -*-

# Having a go at the constraint-based inference algorithm described
#  by Pottier and Remy in ATTPL, chapter 10: "The Essence of ML Type Inference".
#
# Another great reference is a somewhat simplified presentation of the
#  same material, but (thankfully) with some context, by Pottier: "A
#  modern eye on ML type inference - Old techniques and recent
#  developments", available from his home page:
#  http://cristal.inria.fr/~fpottier/
#
# NOTE: this is a work in progress.
# http://www.nightmare.com/rushing/irken
#
# I'm trying *for now* to make this as close to the formal spec as possible,
# in order to understand the algorithm, which makes this Not Very Efficient.
# After that, I plan to do the usual union-find/path-compression tvar stuff,
# and try to grok the 'rank' attribute as described in the chapter.

# TODO: s-letall, letrec, kind-checking

import nodes

from pprint import pprint
import sys
is_a = isinstance

import pdb
trace = pdb.set_trace

# the simply typed lambda calculus:
#   e ::= x | \x.e | e e
# 
# expressions:
# x    : <varref>
# \x.e : <function>
# e e  : <application>

# types
# t ::= a | (arrow t t)
# (where a = <tvar>)

# constraints:
# C ::= (equals t t) | (and C C) | (exists a C)

# types

tvar_counter = 0

def fresh():
    global tvar_counter
    result = tvar_counter
    tvar_counter += 1
    return result

class _type:
    pass

class t_base (_type):
    def __cmp__ (self, other):
        return cmp (self.__class__, other.__class__)

class t_int (t_base):
    def __repr__ (self):
        return 'int'

class t_char (t_base):
    def __repr__ (self):
        return 'char'

class t_str (t_base):
    def __repr__ (self):
        return 'str'

class t_var (_type):
    next = None
    rank = -1
    letters = 'abcdefghijklmnopqrstuvwxyz'
    def __init__ (self):
        self.id = fresh()
    def __repr__ (self):
        if self.id < len(self.letters):
            return self.letters[self.id]
        else:
            return 'T%d' % (self.id - len(self.letters))

class t_predicate (_type):
    def __init__ (self, name, args):
        self.name = name
        self.args = tuple (args)
    def __repr__ (self):
        # special case
        if self.name == 'arrow':
            return '%r->%r' % self.args
        else:
            return '%s%r' % (self.name, self.args)

def is_pred (t, *p):
    # is this a predicate from the set <p>?
    return is_a (t, t_predicate) and t.name in p

def arrow (t0, t1):
    return t_predicate ('arrow', (t0, t1))
    
# row types
def product (row):
    # a.k.a. 'Π'
    # XXX kind-check that args[0] is a row?
    return t_predicate ('product', (row,))

def sum (row):
    # a.k.a. 'Σ'
    return t_predicate ('sum', (row,))

def rdefault (arg):
    # a.k.a. 'δ'
    return t_predicate ('rdefault', (arg,))

def rlabel (name, type, rest):
    return t_predicate ('rlabel', (name, type, rest))

# constraints

def constraint_repr (kind, args):
    if args:
        return '[%s %s]' % (kind, ' '.join ([repr(x) for x in args]))
    else:
        return '%s' % (kind,)

class constraint:
    kind = 'abstract'
    args = ()

    def __repr__ (self):
        return constraint_repr (self.kind, self.args)

class c_true (constraint):
    kind = 'true'
    args = ()

class c_false (constraint):
    kind = 'true'
    args = ()

class c_equals (constraint):
    kind = 'equals'
    def __init__ (self, *args):
        # this class is used for both simple binary equality, but also
        # as a 'multi-equation' for unification, hence the *args
        self.args = args
        if len(args) == 0:
            raise ValueError
    def __repr__ (self):
        return constraint_repr ('=', self.args)
    def split (self):
        # split into tvars and types
        # a multi-equation in standard form will have at most one non-variable.
        tv, ty = [], []
        for e in self.args:
            if is_a (e, t_var):
                tv.append (e)
            else:
                ty.append (e)
        return tv, ty

class c_and (constraint):
    kind = 'and'
    def __init__ (self, c0, c1):
        self.args = (c0, c1)
        
class c_exists (constraint):
    kind = 'exists'
    def __init__ (self, vars, sub):
        self.args = (vars, sub)
        self.vars = vars
        self.sub = sub

class c_is (constraint):
    # <x> has type <t> iff <t> is an instance of the type scheme associated with <x>
    kind = 'is'
    def __init__ (self, x, t):
        self.args = (x, t)
        self.x = x
        self.t = t

class c_let (constraint):
    kind = 'let'
    def __init__ (self, formal, init, body):
        self.args = (formal, init, body)
        self.formal = formal
        self.init = init # this is poorly named, should be 'type' or 'scheme'
        self.body = body

class c_forall (constraint):
    kind = 'forall'
    def __init__ (self, vars, constraint, type):
        self.args = (vars, constraint, type)
        self.vars = vars
        self.constraint = constraint
        self.type = type

# stack frames
class frame:
    kind = 'abstract'
    args = ()

class s_empty (frame):
    kind = 'empty'

empty = s_empty()

class s_and (frame):
    def __init__ (self, c):
        self.args = (c,)
        self.constraint = c

class s_exists (frame):
    def __init__ (self, vars):
        self.args = (vars,)
        self.vars = vars

class s_let (frame):
    def __init__ (self, formal, vars, type, body, rank):
        self.args = (formal, vars, type, body)
        self.formal = formal
        self.vars = vars
        self.type = type
        self.body = body
        for v in vars:
            v.rank = rank

class s_env (frame):
    def __init__ (self, formal, type):
        self.args = (formal, type)
        self.formal = formal
        self.type = type

# this is a two-phase algorithm
# 1) constraint generation
# 2) constraint solving

class constraint_generator:

    def go (self, exp):
        t = t_var()
        return self.gen (exp, t), t

    def gen (self, exp, t):
        if is_a (exp, nodes.varref):
            return c_is (exp.name, t)
        elif is_a (exp, nodes.function):
            assert (len (exp.formals) == 1)
            a1 = t_var()
            a2 = t_var()
            sub0 = self.gen (exp.body, a2)
            let0 = c_let (exp.formals[0].name, a1, sub0)
            # XXX: in ATTPL, this is a c_supertype relation
            sub1 = c_equals (t, arrow (a1, a2))
            return c_exists ((a1, a2), c_and (let0, sub1))
        elif is_a (exp, nodes.application):
            e1 = exp.rator
            # simple lambda calculus only!
            assert (len (exp.rands) == 1)
            e2 = exp.rands[0]
            a = t_var()
            sub0 = self.gen (e1, arrow (a, t))
            sub1 = self.gen (e2, a)
            return c_exists ((a,), c_and (sub0, sub1))
        elif is_a (exp, nodes.let):
            x = t_var()
            assert (len(exp.names) == 1)
            init0 = self.gen (exp.inits[0], x)
            body0 = self.gen (exp.body, t)
            return c_let (exp.names[0].name, c_forall ((x,), init0, x), body0)
        elif is_a (exp, nodes.literal):
            if exp.kind == 'int':
                return c_equals (t, t_int())
            elif exp.kind == 'char':
                return c_equals (t, t_char())
            else:
                raise ValueError ("unsupported literal type")
        else:
            raise ValueError

class UnboundVariable (Exception):
    pass

class Continue (Exception):
    pass

class TypeError (Exception):
    pass

# ok, what kind of interface do we need here?
# we need to be able to add multi-equations to the
# conjunction, and have the unifier automatically
# trigger the rules. pretty sure we'll need a global
# map of tvars as well.

class unifier:
    def __init__ (self, step=False):
        self.step = step
        self.eqs = []
        self.map = {}
        self.exists = []
        self.free = {}

    def add (self, vars, type=None):
        # add a set of terms as a multi-equation to the conjunction.
        # Note: *terms must represent a 'standard' multi-equation.
        assert (is_a (vars, set))
        assert (not is_a (type, t_var))
        if is_a (type, t_predicate):
            args2 = []
            for arg in type.args:
                if is_a (arg, str):
                    # XXX row labels, must be a better way.
                    args2.append (arg)
                elif not is_a (arg, t_var):
                    self.dprint ('s-name-1')
                    x = t_var()
                    self.exists.append (x)
                    self.add (set([x]), arg)
                    args2.append (x)
                else:
                    args2.append (arg)
            type = t_predicate (type.name, args2)
        eq = [vars, type]

        if (not type and len(vars) == 1) or (type and len(vars) == 0):
            self.dprint ('s-single')
        else:
            # only goes into the map if it does *not* fuse.
            fuse = False
            for v in vars:
                if self.map.has_key (v.id):
                    fuse = v
                    break
            if fuse:
                self.fuse (fuse, eq)
            else:
                for v in vars:
                    self.map[v.id] = eq
                    self.inc_free (v)
                self.eqs.append (eq)
                if type:
                    map_type (type, self.inc_free)

    def add2 (self, *args):
        vars = set()
        types = []
        for arg in args:
            if is_a (arg, t_var):
                vars.add (arg)
            else:
                types.append (arg)
        if len(types) == 2:
            self.decompose ((vars, types[0]), (vars, types[1]))
        elif len(types) > 2:
            raise ValueError ("too many types")
        elif len(types) == 1:
            self.add (vars, types[0])
        else:
            self.add (vars, None)

    def inc_free (self, v):
        if is_a (v, t_var):
            if self.free.has_key (v.id):
                self.free[v.id] += 1
            else:
                self.free[v.id] = 1
            
    def dec_free (self, v):
        if is_a (v, t_var):
            self.free[v.id] -= 1

    def forget (self, eq):
        vars, type = eq
        for v in vars:
            del self.map[v.id]
            self.dec_free (v)
        self.eqs.remove (eq)
        if type:
            map_type (type, self.dec_free)

    def compress (self, Y, Z):
        vars, type = eq = self.map[Y.id]
        vars.remove (Y)
        del self.map[Y.id]
        for i in range (len (self.eqs)):
            # substitute it in place
            if self.eqs[i][1]:
                self.eqs[i][1] = subst (Y, Z, self.eqs[i][1])
        self.free[Y.id] = 0
        if (len(vars) == 1 and type is None) or (type and len(vars) == 0):
            self.dprint ('s-single')
            self.forget (eq)
        Y.next = Z

    def fuse (self, v, term):
        other = self.map[v.id]
        assert (term is not other)
        tvs0, ty0 = term
        tvs1, ty1 = other
        self.forget (other)
        self.dprint ('s-fuse')
        if ty0 and ty1:
            # must unify types
            # A=B=T0 ^ B=C=T1 => A=B=C=T0=T1
            self.decompose (term, other)
        else:
            # A=B=T0 ^ B=C => A=B=C=T0
            self.add (tvs0.union (tvs1), ty0 or ty1)

    def decompose (self, t0, t1):
        tvs0, ty0 = t0
        tvs1, ty1 = t1
        tvs = tvs0.union (tvs1)
        if ty0 == ty1:
            # a=b=int=int, etc... => a=b=int
            self.add (tvs, ty0)
        elif is_pred (ty0, 'rlabel', 'rdefault') or is_pred (ty1, 'rlabel', 'rdefault'):
            self.unify_rows (ty0, ty1, tvs)
        elif (is_a (ty0, t_predicate) and is_a (ty1, t_predicate)
              and ty0.name == ty1.name
              and len(ty0.args) == len(ty1.args)):
            self.dprint ('s-decompose')
            # P(a,b,c)=P(d,e,f)=ε => a=d ^ b=e ^ c=f ^ P(a,b,c)=ε
            for i in range (len (ty0.args)):
                self.add2 (ty0.args[i], ty1.args[i])
            self.add (tvs, ty0)
        else:
            self.dprint ('s-clash')
            raise TypeError ((ty0, ty1))

    def unify_rows (self, ty0, ty1, tvs):
        if is_pred (ty0, 'rlabel') and is_pred (ty1, 'rlabel'):
            if ty0.args[0] != ty1.args[0]:
                # distinct head labels
                self.dprint ('s-mutate-ll')
                # XXX be concerned about how one of these may have types
                #     and the other has variables.  do we need to check
                #     and reorder them?
                l0, t0, d0 = ty0.args
                l1, t1, d1 = ty1.args
                x = t_var()
                self.exists.append (x)
                self.add2 (d0, rlabel (l1, t1, x))
                self.add2 (d1, rlabel (l0, t0, x))
                self.add (tvs, rlabel (l0, t0, d0))
            else:
                # XXX this should be handled by the normal s-decompose
                l0, t0, d0 = ty0.args
                l1, t1, d1 = ty1.args
                self.add2 (t0, t1)
                self.add2 (d0, d1)
                self.add (tvs, ty0)
        elif is_pred (ty0, 'rlabel') or is_pred (ty1, 'rlabel'):
            # only one is an rlabel
            if is_pred (ty1, 'rlabel'):
                # ensure that ty0 is the rlabel
                ty0, ty1 = ty1, ty0
            if is_pred (ty1, 'rdefault'):
                self.dprint ('s-mutate-dl')
                x = ty1.args[0]
                assert (is_a (x, t_var))
                self.add2 (x, ty0.args[1])
                self.add2 (ty1, ty0.args[2])
                self.add (tvs, ty1)
            elif is_a (ty1, t_predicate):
                # some other predicate
                self.dprint ('s-mutate-gl')
                n = len (ty1.args)
                tvars0 = [t_var() for x in ty1.args]
                tvars1 = [t_var() for x in ty1.args]
                self.exists.extend (tvars0)
                self.exists.extend (tvars1)
                l0, t0, d0 = ty0.args
                g = ty1.name
                self.add2 (t_predicate (g, tvars0), t0)
                self.add2 (t_predicate (g, tvars1), d0)
                for i in range (n):
                    self.add2 (ty1.args[i], rlabel (l0, tvars0[i], tvars1[i]))
                self.add (tvs, ty1)
            else:
                self.dprint ('s-clash')
                raise TypeError ((ty0, ty1))
        elif is_pred (ty0, 'rdefault',) or is_pred (ty1, 'rdefault'):
            if is_pred (ty1, 'rdefault'):
                # ensure that ty0 is the rdefault/δ
                ty0, ty1 = ty1, ty0
            if is_a (ty1, t_predicate):
                # some other predicate
                self.dprint ('s-mutate-gd')
                n = len (ty1.args)
                g = ty1.name
                tvars = [ t_var() for x in ty1.args ]
                self.exists.extend (tvars)
                self.add2 (ty0.args[0], t_predicate (g, tvars))
                for i in range (n):
                    self.add2 (ty1.args[i], rdefault (tvars[i]))
                self.add (tvs, ty0)
            else:
                self.dprint ('s-clash')
                raise TypeError ((ty0, ty1))
        else:
            self.dprint ('s-clash')
            raise TypeError ((ty0, ty1))

    def split (self, sz):
        # we leave in only equations made entirely of 'old' variables
        young = set (sz.vars)
        u2 = []
        forget = []
        for eq in self.eqs:
            [vars, type] = eq
            if young.intersection (vars) or (type and set (ftv (type)).intersection (vars)):
                vars, type = eq
                u2.append (c_equals (* list(vars) + [type]))
                forget.append (eq)
        for eq in forget:
            self.forget (eq)            
        return list_to_conj (u2)

    def dprint (self, msg):
        if self.step:
            sys.stderr.write ('*** ')
            sys.stderr.write (msg)
            sys.stderr.write ('\n')
            self.pprint()

    def pprint (self):
        sys.stdout.write ('U: ')
        for vars, type in self.eqs:
            #sys.stdout.write ('='.join (['%r.%d' % (v,v.rank) for v in vars]))
            sys.stdout.write ('='.join (['%r' % v for v in vars]))
            if type:
                sys.stdout.write ('=%r' % (type,))
            sys.stdout.write (' ')
        sys.stdout.write ('\n')

# I believe the 'union-find' data-structure/algorithm is another way
#   of describing HM's use of type variable lookup with path
#   compression.

class solver:

    def __init__ (self, step=True):
        self.step = step
        self.prim_env = self.make_prim_env()

    def make_prim_env (self):
        # build type schemes for builtins/primitives
        int = t_int()
        char = t_char()
        e = {}
        e['%+'] = arrow (int, arrow (int, int))
        e['%-'] = arrow (int, arrow (int, int))
        # XXX consider generating the tvars fresh every time
        x = t_var()
        y = t_var()
        e['%make-pair'] = c_forall ((x,y), c_true(), arrow (x, arrow (y, t_predicate ('pair', (x, y)))))
        e['%pair/first'] = c_forall ((x,y), c_true(), arrow (t_predicate ('pair', (x, y)), x))
        e['%pair/second'] = c_forall ((x,y), c_true(), arrow (t_predicate ('pair', (x, y)), y))
        e['%char->int'] = arrow (char, int)
        e['%int->char'] = arrow (int, char)
        # rows
        e['%make-row'] = c_forall ((x,), c_true(), arrow (x, product (rdefault (x))))
        # label-related row types are in the 'lookup_special' method
        return e

    def dprint (self, msg):
        if self.step:
            sys.stderr.write (msg)
            sys.stderr.write ('\n')

    def solve (self, c):

        self.dprint ('\nHit <return> at each pause (or "t<return>" to enter the debugger)')

        pvars = []
        exists = []
        # ensure there are always two items on the stack
        s = [empty, empty]
        u = unifier (self.step)
        c = c
        orig_c = c
        rank = 0

        def push (x):
            s.append (x)

        def pop ():
            s.pop()

        while 1:

            if self.step:
                print 'S:',
                self.pprint_stack (s)
                u.pprint()
                print 'C:', c
                print 'exists:', exists

            # the top two elements of the stack
            sy, sz = s[-2], s[-1]

            if self.step:
                print '-----------------------------'
                if raw_input().startswith ('t'):
                    trace()

            # --- solver ---            

            # XXX everywhere here, alpha-conversion to avoid capture (i.e., all the X!=ftv(e) rules)
            if u.exists:
                self.dprint ('s-ex-1')
                push (s_exists (tuple (u.exists)))
                u.exists = []
            elif is_a (sy, s_and) and is_a (sz, s_exists):
                self.dprint ('s-ex-2')
                # swap 'em
                pop(); pop()
                push (sz); push (sy)
            elif is_a (sy, s_let) and is_a (sz, s_exists):
                self.dprint ('s-ex-3')
                pop(); pop()
                vars = sy.vars + sz.vars
                push (s_let (sy.formal, sy.vars + sz.vars, sy.type, sy.body, rank))
            elif is_a (sy, s_env) and is_a (sz, s_exists):
                self.dprint ('s-ex-4')
                pop(); pop()
                push (sz)
                push (sy)
            elif is_a (sy, s_exists) and is_a (sz, s_exists):
                # XXX hack?
                self.dprint ('s-ex-5')
                pop(); pop();
                push (s_exists (sy.vars + sz.vars))
            elif is_a (sz, s_exists):
                self.dprint ('s-record-ex')
                exists.extend (sz.vars)
                pop()
            elif is_a (c, c_equals):
                self.dprint ('s-solve-eq')
                u.add2 (*c.args)
                c = c_true()
            elif is_a (c, c_is) and is_a (c.x, str):
                self.dprint ('s-solve-id')
                scheme, type = self.lookup (c.x, s), c.t
                # assert that scheme.type is a tvar
                # "Recall that if σ is of the form ∀X0..XN[U].X
                # where X0..XN#ftv(T), then c_is(σ, T) stands for ∃X0..XN.(U ^ X=T)."
                self.dprint ('scheme= %r' % scheme)
                self.dprint ('type=%r' % type)
                if is_a (scheme, c_forall):
                    if not scheme.vars and is_a (scheme.constraint, c_true):
                        c = c_equals (scheme.type, type)
                    else:
                        c = c_exists (scheme.vars, c_and (scheme.constraint, c_equals (scheme.type, type)))
                else:
                    c = c_equals (scheme, type)
            elif is_a (c, c_and):
                self.dprint ('s-solve-and')
                push (s_and (c.args[1]))
                c = c.args[0]
            elif is_a (c, c_exists):
                self.dprint ('s-solve-ex')
                push (s_exists (c.vars))
                c = c.sub
            elif is_a (c, c_let):
                self.dprint ('s-solve-let')
                if is_a (c.init, c_forall):
                    vars = c.init.vars
                    push (s_let (c.formal, c.init.vars, c.init.type, c.body, rank))
                    rank += 1
                    c = c.init.constraint
                else:
                    # let x: T in C == let x: ∀∅[true].T in C
                    push (s_let (c.formal, (), c.init, c.body, rank))
                    c = c_true()
            elif is_a (c, c_true):
                if is_a (sz, s_and):
                    self.dprint ('s-pop-and')
                    pop()
                    c = sz.constraint
                elif is_a (sz, s_let) and not is_a (sz.type, t_var):
                    self.dprint ('s-name-2')
                    x = t_var()
                    pop()
                    push (s_let (sz.formal, sz.vars + (x,), x, sz.body, rank))
                    u.add (set([x]), sz.type)
                elif is_a (sz, s_let):
                    # s-compress, s-unname, s-letall, s-pop-let
                    # these rules all require an examination of the variables
                    #  bound by this <let>.
                    try:
                        for var in sz.vars:
                            if u.free[var.id] == 0 and sz.type is not var:
                                # completely unused variable
                                self.dprint ('s-unname-2 (Y=%r)' % (var,))
                                pop()
                                vars = tuple ([x for x in sz.vars if x is not var])
                                push (s_let (sz.formal, vars, sz.type, sz.body, rank))
                                raise Continue
                            
                            if u.map.has_key (var.id):
                                # for s-compress, we want to know if there's a multi-equation
                                # of the form Y=Z=ε (where Y is bound by this <let>)
                                vars, ty = u.map[var.id]
                                if len(vars) > 1:
                                    self.dprint ('s-compress')
                                    lv = list(vars)
                                    if lv[0] is var:
                                        Z = lv[1]
                                    else:
                                        Z = lv[0]
                                    u.compress (var, Z)
                                    self.dprint ('  Y=%r Z=%r' % (var, Z))
                                    szvars = tuple([x for x in sz.vars if x is not var])
                                    if sz.type == var:
                                        sztype = Z
                                    else:
                                        sztype = sz.type
                                    pop()
                                    push (s_let (sz.formal, szvars, sztype, sz.body, rank))
                                    raise Continue
                                elif u.free[var.id] == 1 and var != sz.type and len(u.map[var.id][0]) > 1:
                                    szvars = tuple([x for x in sz.vars if x is not var])
                                    self.dprint ('s-unname')
                                    self.dprint (' Y=%r' % (var,))
                                    pop()
                                    push (s_let (sz.formal, szvars, sz.type, sz.body, rank))
                                    evars, etype = u.map[var.id]
                                    evars.remove (var)
                                    raise Continue
                        #elif s-letall:
                        # s-letall will simplify the current scheme by removing some of
                        #   the quantified tvars.
                        else:
                            # pop-let is a fall-through - after all
                            # the above conditions have been met it turns the <let>
                            # into an <env>.
                            self.dprint ('s-pop-let')
                            u2 = u.split (sz)
                            #u.pprint()
                            #print 'u2=%r' % (u2,)
                            pop()
                            push (s_env (sz.formal, c_forall (sz.vars, u2, sz.type)))
                            c = sz.body
                    except Continue:
                        # back to the top of the solver loop
                        pass
                elif is_a (sz, s_env):
                    pvars.append ((sz.formal, sz.type.type))
                    self.dprint ('s-pop-env')
                    pop()
                elif is_a (sz, s_empty):
                    self.dprint ('exists=%r' % exists)
                    self.dprint ('constraint=%r' % orig_c)
                    return pvars, u
                else:
                    raise ValueError ("unexpected")
            else:
                raise ValueError ("no rule applies")

    def lookup (self, x, s):
        # XXX side conditions
        for f in s:
            if is_a (f, s_and):
                continue
            elif is_a (f, s_exists):
                continue
            elif is_a (f, s_let):
                if f.formal == x:
                    raise ValueError ("shouldn't happen?")
                continue
            elif is_a (f, s_env):
                if f.formal != x:
                    continue
                else:
                    return f.type
        else:
            return self.lookup_special_names (x)

    def lookup_special_names (self, name):
        # we need this hack for row-related label lookups, since the label
        #  is part of the name.
        if self.prim_env.has_key (name):
            return self.prim_env[name]
        elif name.startswith ('%rextend/'):
            what, label = name.split ('/')
            x, y, z = t_var(), t_var(), t_var()
            # ∀XYZ.Π(l:X;Y) → Z → Π(l:Z;Y)
            return c_forall (
                (x,y,z),
                c_true(),
                arrow (product (rlabel (label, x, y)),
                       arrow (z, product (rlabel (label, z, y))))
                )
        elif name.startswith ('%raccess/'):
            what, label = name.split ('/')
            x, y = t_var(), t_var()
            # ∀XY.Π(l:X;Y) → X
            return c_forall ((x, y), c_true(), arrow (product (rlabel (label, x, y)), x))
        else:
            raise UnboundVariable (name)

    def pprint_stack (self, s):

        W = sys.stdout.write

        W ('\n')
        n = len(s)
        # the 2 is for the two <empty> sentinels
        for i in range (2,n):
            W ('%2d: ' % (i-2,))
            si = s[i]
            if is_a (si, s_empty):
                W ('[]')
            elif is_a (si, s_and):
                W ('[] ^ %s' % si.args)
            elif is_a (si, s_exists):
                W ('exists%r.[]' % si.args)
            elif is_a (si, s_let):
                W ('let %s: forall%r[[]].%r in %r' % si.args)
            elif is_a (si, s_env):
                W ('env %s: %r in []' % si.args)
            else:
                raise NotImplementedError
            W ('\n')

def ftv (e):
    r = set()
    def f (e):
        if is_a (e, t_var):
            r.add (e)
        elif is_a (e, t_predicate):
            for arg in e.args:
                f (arg)
        elif is_a (e, t_base):
            pass
        elif is_a (e, str):
            # XXX row labels
            pass
        else:
            raise ValueError ("unknown type")
    f (e)
    return r

# need a 'map_constraint' that will map a function over any constraint.
def subst (v0, v1, e):
    # replace all <v0> with <v1>
    def f (e):
        if is_a (e, t_var):
            if e.id == v0.id:
                return v1
            else:
                return e
        elif is_a (e, t_predicate):
            return t_predicate (e.name, [f(x) for x in e.args])
        if is_a (e, c_equals):
            return c_equals (*[f(x) for x in e.args])
        elif is_a (e, c_and):
            return c_and (*[f(x) for x in e.args])
        elif is_a (e, t_base):
            return e
        elif is_a (e, str):
            # XXX fixme - this happens with row labels, might need their own predicate?
            return e
        else:
            raise ValueError ("unknown expression in subst")
    r = f (e)
    return r

def list_to_conj (l):
    if len(l) == 0:
        return c_true()
    elif len(l) == 1:
        return l[0]
    else:
        r = l[0]
        for x in l[1:]:
            r = c_and (r, x)
        return r

def map_type (t, f):
    if is_a (t, t_var):
        return f (t)
    elif is_a (t, t_predicate):
        return t_predicate (t.name, [f(x) for x in t.args])
    elif is_a (t, t_base):
        return f (t)
    elif is_a (t, str):
        return t
    else:
        raise ValueError ("unknown type object")

def print_solution (pvars, u, top_tv):

    print pvars
    u.pprint()
    print top_tv

    def lookup (x):
        if is_a (x, t_var):
            if x.next:
                # follow path compression
                return lookup (x.next)
            elif u.map.has_key (x.id):
                vars, type = u.map[x.id]
                return lookup (type)
            else:
                return x
        elif is_a (x, t_predicate):
            return t_predicate (x.name, [lookup(y) for y in x.args])
        elif is_a (x, t_base):
            return x
        elif is_a (x, str):
            # XXX row labels
            return x
        else:
            raise ValueError ("unknown type object")

    for pvar, tvar in pvars:
        print '%r: %r' % (pvar, lookup (tvar))

    print 'program: %r' % lookup (top_tv)

def read_string (s):
    import cStringIO
    import lisp_reader
    sf = cStringIO.StringIO (s)
    r = lisp_reader.reader (sf)
    return r.read()

def test (s, step=True):
    import nodes
    from pprint import pprint as pp
    global tvar_counter
    tvar_counter = 0
    # wrap everything in a top-level <let>
    # if we omit this, s-compress won't work on top-level exists
    print 'expression=', s
    s = "(let ((top %s)) top)" % s
    exp = read_string (s)
    w = nodes.walker()
    exp2 = w.go (exp)
    # alpha conversion
    nodes.rename_variables (exp2)
    cg = constraint_generator()
    c, top_tv = cg.go (exp2)
    print 'constraint=', c
    m, u = solver(step).solve(c)
    print_solution (m, u, top_tv)

tests = [
    "5",
    "(lambda (x) 3)",
    "(lambda (x) x)",
    "(let ((f (lambda (x) x))) (f 5))",
    "(lambda (x) (lambda (y) x))",
    "(let ((f (lambda (x) x))) f)",
    "((lambda (x) x) (lambda (x) x))",
    "(%+ 3)",
    "((%+ 3) 4)",
    "((%- ((%+ 3) 4)) 2)",
    "((%make-pair 3) 4)",
    "((%make-pair 3) #\\A)",
    "(%pair/first ((%make-pair 3) #\\A))",
    "(%pair/second ((%make-pair 3) #\\A))",
    "(%char->int #\\A)",
    "(%int->char 65)",
    # this causes a cycle, breaks print-solution
    #"(let ((f (lambda (x) 3))) (%int->char (f f)))",
    # row tests
    # make a default row - maps all symbols to a function: δ(a->a)
    "(%make-row (lambda (x) x))",
    # lookup via a random field name on a row of type δ(int)
    "(%raccess/field (%make-row 0))",
    # extend a row: (label:char;δ(int))
    "((%rextend/label (%make-row 0)) #\\A)",
    # dereference via the defined label
    "(%raccess/label ((%rextend/label (%make-row 0)) #\\A))",
    # dereference via an undefined label
    "(%raccess/fnord ((%rextend/label (%make-row 0)) #\\A))",
    # extend a row twice: (other:a->a; label:char; δ(int))
    "((%rextend/other ((%rextend/label (%make-row 0)) #\\A))(lambda (x) x))",
    # deref one defined label
    "(%raccess/label ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # deref the other
    "(%raccess/other ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # deref an undefined label
    "(%raccess/fnord ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # build an extended record, store it in a variable, then deref to pull the identity function out and use it.
"""(let ((rec0 ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x))))
     (let ((f0 (%raccess/other rec0)))
       (f0 9)))""",
    ]

if __name__ == '__main__':
    if '-t' in sys.argv:
        for t in tests:
            test (t, step=False)
    elif '-p' in sys.argv:
        import profile
        profile.run ("test (tests[-1], step=False)")
    elif '-i' in sys.argv:
        # interactive
        while 1:
            sys.stdout.write ('> ')
            line = raw_input()
            if not line:
                break
            else:
                #test (line, step=False)
                test (line, step=True)
    else:
        test ("(lambda (x) x)")
